//**********************************************************************************
// UNIT:    MGDBSessionCache
// FILE:    TMGDBSessionCache.pas
// AUTHOR:  Martin Geier
// VERSION: 1.0
//
// PURPOSE:
//   TMGDBSessionCache caches database sessions along with a TDatabase component.
//   TMGDBSessionCache was written for 3-Tier database applications where
//   normally each new thread in the middle tier has to create a new session and
//   connect to the database. Connecting to the database, especially to SQL servers,
//   takes a certain amount of time. If an application constists of several remote
//   datamodules that each serve one client a lot of connecting/disconnecting
//   from the database is necessary.
//   MGDBSessionCache solves this problem by creating a certain amount of sessions
//   along with TDatabase components attached to the sessions at program start.
//   Threads can request a session from TMGDBSessionCache and will get the session
//   name and the database name to be used in return. The returned session is already
//   connected to the database and ready to use. When a thread has done its database
//   jobs and no longer needs a session, the thread releases the session so
//   TMGDBSessionCache can reuse it.
//   An instance of TMGDBSessionCache can connect to only one database, so if an
//   application needs to connect to different databases, multiple TMGDBSessionCaches
//   have to be used.
//   TMGDBSessionCache is completly thread save. All methods are protected by
//   one critical section, i.e. session requests and releases will by serialized.
//   Using only one critical section may be a little bit restrictive but
//   it is save and prevents deadlocks. Calls to RequestSession and ReleaseSession
//   are very short anyway.
//   I'm using this component in the middle tier of internet applications.
//   I write my one ASP components that use MIDAS to connect to the application
//   server where TMGDBSessionCache is used to connect to the database server.
//   The speed gain to create an HTML page that needs 5 database connections
//   is huge. Page creation is almost instantaneously with TMDGDBSessionCache
//   where without it takes several seconds using Oracle as database server.
//
//
// PROPERTIES:
//   DBNamePrefix, SessionNamePrefix: TMGDBSessionCache uses these properties
//     to name sessions and databases when creating them. To make the names unique
//     an incrementing number is added by TMGDBSessionCache. A pair, session and
//     database, will always have the same number. When using multiple TMGDBSessionCache
//     within one application make sure to set DBNamePrefix and SessionNamePrefix different
//     for each TMGDBSessionCache otherwise the names will collide.
//
//   DBAliasName, DBReadOnly, DBParams, DBTransIsolation: These properties
//     are used when creating new TDatabase components.  DBAliasName and
//     DBParams must be set so that TMGDBSessionCache can connect to the database.
//     DBParams must contain valid USER NAME and PASSWORD settings for the database.
//
//   SessionNetFileDir, SessionPrivateDir: Set when using Paradox as database engine.
//
//   MinSessions: Minimum number of sessions. When calling Connect MinSessions number
//     of database connections will be created by TMGDBSessionCache. When calling
//     CollectGarbage at least MinSessions sessions will remain active.
//     Decreasing MinSessions after a call to Connect will have no effect without
//     calling CollectGarbage. By increasing MinSessions TMGDBSessionsCache will make
//     sure MinSessions sessions are available and if necessary create new sessions.
//
//   MaxSessions: Specifies the maximum number of sessions TMGDBSessionCache can create.
//     Decreasing MaxSessions will never close open sessions, i.e. SessionsCount could
//     be greater than MaxSessions. Increasing MaxSessions will never create new sessions.
//
//   SessionsCount: Read-only; gives the number of database connections TMGDBSessionCache
//     has established.
//
//   SessionsUsed: Read-only; gives the number of sessions currently being used.
//     SessionsCount - SessionsUsed therefore gives the number sessions available.
//
//   SessionsMaxUsed: Read-only; gives the maximum number of sessions used at the
//     same time. SessionsMaxUsed can be used to determine whether it is necessary to
//     change MaxSessions.
//
//   AllowNewSessions: If False, TMGDBSessionCache will not create new sessions. If
//     there are not enough sessions to satisfy the requests an exception will be
//     raised.
//
//   Changing DBNamePrefix, SessionNamePrefix, DBReadOnly, FDBTransIsolation,
//   SessionNetFileDir, SessionPrivateDir, SessionAliasName and SessionDBParams
//   will only affect new sessions. Already open sessions will not be changed.
//
//
// METHODS:
//   Create(AOwner: TComponent): Creates the component but does not create any
//     sessions nor connect to the database.
//
//  Connect: Creates MinSessions connections to the database. Connect should be
//     called during application initialization to start the session cache. If
//     connect is not called, TMGDBSessionCache will not create any sessions until
//     the first session is requested. The MinSession property will not be used
//     in this case.
//
//  CloseAllSessions: Closes all database connections regardless if they are in use
//     or not. Make sure no thread uses a database connection from TMGDBSessionsCache
//     after a call to CloseAllSessions since the connection will be invalid.
//     It is best to make sure SessionsUsed is 0 and to set AllowNewSessions to False
//     before calling CloseAllSessions. CloseAllSessions is automatically called ny
//     the destructor of TMGDBSessionCache.
//
//  RequestSession(var DBName: String; var SessionName: String): Used by threads to
//     request a database session. If successful RequestSession returns the database
//     name and the session name to be used. If no session if available and no new session
//     could be created an exception is raised.
//
//  RequestSession(DS: TDBDataSet): Used by threads to request a database session.
//     If successful DS.DatabaseName and DS.SessionName are set to the database name
//     and the session name to be used. If no session if available and no new session
//     could be created an exception is raised.
//
//  ReleaseSession(SessionName: String): To be called by threads to release a previoulsy
//     requested session. Make sure all threads call ReleaseSession as soon as they no
//     longer need the database connection. This will garantuee the best utilization of
//     the available sessions.
//
//  CollectGarbage: Tries to destroy as many sessions as possible but leaves at least
//     MinSessions sessions open. Sessions currently being used will not be destroyed.
//     CollectGarbage is useful if after a heavy load period, when a lot of new sessions
//     were created to satisfy demand, the number of sessions should be reduced since
//     the sessions are no longer needed. CollectGarbage is the only way to reduce the
//     number of database connections TMGDBSessionsCache retains other than CloseAllSessions.
//
//
// EVENTS:
//   OnNewSession: Fired after a new session has been created.
//   OnRequestSession: Fired when a session is requested.
//   OnReleaseSession: Fired when a session is released.
//   OnCollectGarbage: Fired after garbage collection. 
//
//
// HISTORY:
//   2/6/00 Created
//**********************************************************************************
unit MGDBSessionCache;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, DB,
  DBTables, Syncobjs;

type
	// forward declarations
  TMGDBSessionCache = class;

  // exceptions
  ECannotCreateSession = class(Exception);
  ESessionUnknown = class(Exception);

  // class for cache items
	TCacheItem = class
  private
  	FSession: TSession;
    FDB: TDatabase;
    FUsed: Boolean;

  public
    constructor Create;
    destructor Destroy; override;
  	procedure Clear;
    procedure CloseSession;
    procedure SetSession(DB: TDatabase; S: TSession);
    function GetSession: TSession;
    function GetDatabase: TDatabase;
    procedure UseSession;
    procedure ReleaseSession;
    function IsUsed: Boolean;
    function IsOpen: Boolean;
  end;

  TMGDBSessionCache = class(TComponent)
  private
    FCritSection: TCriticalSection;		// used to synchronize requests from different threads
    FDBNamePrefix: String;
    FDBAliasName: String;
    FSessionNamePrefix: String;
    FDBReadOnly: Boolean;
    FDBParams: TStrings;
    FDBTransIsolation: TTransIsolation;
    FAllowNewSessions: Boolean;
    FMaxSessions: Integer;
    FMinSessions: Integer;
    FSessionsUsed: Integer;
    FSessionsMaxUsed: Integer;
    FSessionsCount: Integer;
    FSessionNetFileDir: String;
    FSessionPrivateDir: String;
    FOnNewSession: TNotifyEvent;
    FOnCollectGarbage: TNotifyEvent;
    FOnRequestSession: TNotifyEvent;
  	FCacheItems: array of TCacheItem;
    FOnReleaseSession: TNotifyEvent;

    // property set methods
    procedure SetDBParams(const Value: TStrings);
    procedure SetMaxSessions(const Value: Integer);
    procedure SetMinSessions(const Value: Integer);
    procedure SetOnCollectGarbage(const Value: TNotifyEvent);
    procedure SetOnNewSession(const Value: TNotifyEvent);
    procedure SetOnRequestSession(const Value: TNotifyEvent);
    procedure SetOnReleaseSession(const Value: TNotifyEvent);

  protected
  	// helper methods
    function NewSession: Integer; virtual;
   	function FindFreeSession: Integer; virtual;

    // event handler procedures
    procedure EvOnNewSession; virtual;
    procedure EvOnRequestSession; virtual;
    procedure EvOnReleaseSession; virtual;
    procedure EvOnCollectGarbage; virtual;

  public
  	// construction, initialization, destruction
  	constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure Connect; virtual;

    // session handling
    procedure RequestSession(var DBName: String; var SessionName: String);  overload; virtual;
    procedure RequestSession(DS: TDBDataSet); overload; virtual;
    procedure ReleaseSession(SessionName: String);

    // cleanup
    procedure CollectGarbage; virtual;
    procedure CloseAllSessions; virtual;


  published
  	// properties
  	property DBAliasName: String read FDBAliasName write FDBAliasName;
    property DBNamePrefix: String read FDBNamePrefix write FDBNamePrefix;
    property SessionNamePrefix: String read FSessionNamePrefix write FSessionNamePrefix;
    property DBReadOnly: Boolean read FDBReadOnly write FDBReadOnly;
    property DBParams: TStrings read FDBParams write SetDBParams;
    property DBTransIsolation: TTransIsolation read FDBTransIsolation write FDBTransIsolation;
    property SessionNetFileDir: String read FSessionNetFileDir write FSessionNetFileDir;
    property SessionPrivateDir: String read FSessionPrivateDir write FSessionPrivateDir;
    property MinSessions: Integer read FMinSessions write SetMinSessions;
    property MaxSessions: Integer read FMaxSessions write SetMaxSessions;
    property SessionsCount: Integer read FSessionsCount;
    property SessionsUsed: Integer read FSessionsUsed;
    property SessionsMaxUsed: Integer read FSessionsMaxUsed;
    property AllowNewSessions: Boolean read FAllowNewSessions write FAllowNewSessions;

    // event properties
		property OnNewSession: TNotifyEvent read FOnNewSession write SetOnNewSession;
    property OnRequestSession: TNotifyEvent read FOnRequestSession write SetOnRequestSession;
    property OnReleaseSession: TNotifyEvent read FOnReleaseSession write SetOnReleaseSession;
    property OnCollectGarbage: TNotifyEvent read FOnCollectGarbage write SetOnCollectGarbage;
  end;

procedure Register;

implementation

uses Math;

const
	// default property values
	DEF_MIN_SESSIONS = 1;
  DEF_MAX_SESSIONS = 50;
  DEF_DBNAME_PREFIX = 'CacheDatabase';
  DEF_SESSIONNAME = 'CacheSession';


//----------------------------------------------------------------------------------
// TCacheItem.Create
//
// Author: Martin Geier
// Creation date: 2/7/00
//----------------------------------------------------------------------------------
constructor TCacheItem.Create;
begin
	inherited Create;

  Clear;
end;

//----------------------------------------------------------------------------------
// TCacheItem.Destroy
//
// Author: Martin Geier
// Creation date: 2/7/00
//----------------------------------------------------------------------------------
destructor TCacheItem.Destroy;
begin
	CloseSession;

  if assigned(FDB) then
  	FDB.Free;

  inherited Destroy;
end;

//----------------------------------------------------------------------------------
// TCacheItem.Clear
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TCacheItem.Clear;
begin
  FSession := nil;
  FDB := nil;
  FUsed := False;
end;

//----------------------------------------------------------------------------------
// TCacheItem.CloseSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TCacheItem.CloseSession;
begin
	if not assigned(FSession) then
  	raise ESessionUnknown.Create('Cannot close not existing session');
	FSession.Close;
	FSession := nil;
  if assigned(FDB) then
  	FDB.Free;
  FUsed := False;
end;

//----------------------------------------------------------------------------------
// TCacheItem.IsUsed
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
function TCacheItem.IsUsed: Boolean;
begin
	Result := FUsed;
end;

//----------------------------------------------------------------------------------
// TCacheItem.IsOpen
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
function TCacheItem.IsOpen: Boolean;
begin
	Result := assigned(FSession);
end;

//----------------------------------------------------------------------------------
// TCacheItem.ReleaseSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TCacheItem.ReleaseSession;
begin
	try
		if not FUsed then
	  	raise Exception.Create('Releasing a not used session');
  finally
	  FUsed := False;
  end;
end;

//----------------------------------------------------------------------------------
// TCacheItem.SetSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TCacheItem.SetSession(DB: TDatabase; S: TSession);
begin
	FSession := S;
  FDB := DB;
end;

//----------------------------------------------------------------------------------
// TCacheItem.GetSession
//
// Author: Martin Geier
// Creation date: 2/7/00
//----------------------------------------------------------------------------------
function TCacheItem.GetSession: TSession;
begin
	if IsOpen then
		Result := FSession
  else
  	raise Exception.Create('No session in cache item');
end;

//----------------------------------------------------------------------------------
// TCacheItem.GetDatabase
//
// Author: Martin Geier
// Creation date: 2/7/00
//----------------------------------------------------------------------------------
function TCacheItem.GetDatabase: TDatabase;
begin
	if IsOpen then
		Result := FDB
  else
  	raise Exception.Create('No session in cache item');
end;

//----------------------------------------------------------------------------------
// TCacheItem.UseSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TCacheItem.UseSession;
begin
	if not IsOpen then
  	raise ESessionUnknown.Create('Cannot use a non existing session');
  FUsed := True;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.Create
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
constructor TMGDBSessionCache.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

	// initialize properties
  FDBNamePrefix := DEF_DBNAME_PREFIX;
  FDBAliasName := '';
  FSessionNamePrefix := DEF_SESSIONNAME;
  FDBReadOnly := False;
  FDBTransIsolation := tiReadCommitted;
  FAllowNewSessions := True;
  FMaxSessions := DEF_MAX_SESSIONS;
  FMinSessions := DEF_MIN_SESSIONS;
  FSessionsUsed := 0;
  FSessionsMaxUsed := 0;
  FSessionsCount := 0;
  FSessionNetFileDir := '';
  FSessionPrivateDir := '';
  FDBParams := TStringList.Create;

  // create critical section
	FCritSection := TCriticalSection.Create;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.Destroy
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
destructor TMGDBSessionCache.Destroy;
begin
	// close all open sessions
	CloseAllSessions;
  
	FCacheItems := nil;
	if assigned(FDBParams) then
  	FDBParams.Free;

	if assigned(FCritSection) then
  	FCritSection.Free;

  inherited Destroy;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.Loaded
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.Loaded;
var
	i: Integer;
begin
  // initialize FUsedSessions
	SetLength(FCacheItems, FMaxSessions);
  for i := 0 to Length(FCacheItems) - 1 do
  	FCacheItems[i] := TCacheItem.Create;

	inherited Loaded;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.Connect
//
// Start as many sessions as specified in MinSessions. This procedure should be
// called only once at program start.
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.Connect;
begin
  // open MinSession sessions
  SetMinSessions(FMinSessions);
end;


//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetDBParams
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetDBParams(const Value: TStrings);
begin
  FDBParams.Assign(Value);
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetMaxSessions
//
// Set new maximum number of sessions. If we have more open sessions then
// the new value we set MaxSessions to SessionsCount.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetMaxSessions(const Value: Integer);
begin
	try
  	FCritSection.Enter;

    // if MaxSessions is less than MinSessions we raise an exception
    if Value < FMinSessions then
      raise Exception.Create('MaxSessions cannot be less than MinSessions');

    FMaxSessions := Max(Value, SessionsCount);

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetMinSessions
//
// Set new minimum number of sessions. If Value is less than 0, MinSessions
// is set to 0. If Value is greater than SessionsCount we create the necessary
// new sessions.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetMinSessions(const Value: Integer);
var
	i: Integer;
begin

	try
  	FCritSection.Enter;

    // if MinSessions is greater than MaxSessions we raise an exception
    if Value > FMaxSessions then
      raise Exception.Create('MinSessions cannot be greater than MaxSessions');

    // if we are in design mode we just set the property but do not create
    // any sessions
    if (csDesigning in ComponentState) or (csReading in ComponentState) then
    begin
      FMinSessions := Max(0, Value);
      exit;
    end;

    if Value < 0 then
      FMinSessions := 0
    else if Value <= SessionsCount then
      FMinSessions := Value
    else
    begin
      for i := SessionsCount + 1 to Value do
        NewSession;
      FMinSessions := Value;
    end;

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.CloseAllSessions
//
// Destroy all sessions regardless whether they are used or not.
//
// ATTENTION: Make sure no dataset reuses a sessions after CloseAllSessions
// requested before the call of CloseAllSession.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.CloseAllSessions;
var
	i: Integer;
begin
	try
  	FCritSection.Enter;

    // loop over all cache entries
    for i := 0 to Length(FCacheItems) - 1 do
      if FCacheItems[i].IsOpen then
        FCacheItems[i].CloseSession;
    FSessionsCount := 0;
    FSessionsUsed := 0;

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.CollectGarbage
//
// Closes all not in use sessions but leave at least MinSessions sessions
// active.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.CollectGarbage;
var
	i: Integer;
begin
	try
  	FCritSection.Enter;

    // loop over all cache entries
    for i := Length(FCacheItems) - 1 downto 0 do
    begin
    	// stop when MinSessions sessions is reached
      if FSessionsCount <= FMinSessions then
        break;

      // close open but not used sessions
      if FCacheItems[i].IsOpen and not FCacheItems[i].IsUsed then
      begin
        FCacheItems[i].CloseSession;
        FSessionsCount := FSessionsCount - 1;
      end;
    end;

    // trigger OnGarbageCollect event
    EvOnCollectGarbage;

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.RequestSession
//
// Try to find a not used session and use it. If none can be found create a
// new session and use it.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.RequestSession(var DBName: String; var SessionName: String);
var
	i: Integer;
begin
	try
  	FCritSection.Enter;

    // try to find a free session
    i := FindFreeSession;

    // if there are no sessions available, create a new one
    if i < 0 then
      i := NewSession;

    // set session state to used
    FCacheItems[i].UseSession;
    FSessionsUsed := FSessionsUsed + 1;
    FSessionsMaxUsed := Max(FSessionsUsed, FSessionsMaxUsed);

    // get DBName and SessionName
    DBName := FCacheItems[i].GetDatabase.DatabaseName;
    SessionName := FCacheItems[i].GetSession.SessionName;

    // trigger OnRequestSession event
    EvOnRequestSession;

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.RequestSession
//
// Call above RequestSession to get a new session and assign DBName and
// SessionName to DS.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.RequestSession(DS: TDBDataSet);
var
	DBName, SessionName: String;
begin
   RequestSession(DBName, SessionName);
   DS.DatabaseName := DBName;
   DS.SessionName := SessionName;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.ReleaseSession
//
// Sets the sessions to not used so it is available for another request.
// If the session cannot be found or the sessions is not currently used
// an exception is raised.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.ReleaseSession(SessionName: String);
var
	i: Integer;
begin
	try
  	FCritSection.Enter;

    // search for the session
    for i := 0 to Length(FCacheItems) - 1 do
      if FCacheItems[i].IsOpen and (FCacheItems[i].GetSession.SessionName = SessionName) then
      begin
        FCacheItems[i].ReleaseSession;	// we found the session, release it
        FSessionsUsed := FSessionsUsed - 1;

        // trigger OnReleaseSession event
        EvOnReleaseSession;

        exit;
      end;

    // if we reach this point, there is no session with name SessionName
    raise ESessionUnknown.Create('Session ' + SessionName + ' not found');

  finally
  	FCritSection.Leave;
  end;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.FindFreeSession
//
// Tries to find an open, not used sessions. If one can be found, the index
// of the session in FCacheItems is returned otherwise -1 is returned.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
function TMGDBSessionCache.FindFreeSession: Integer;
var
	i: Integer;
begin
	// search for a session that is open and not used
  for i := 0 to Length(FCacheItems) - 1 do
  	if FCacheItems[i].IsOpen and not FCacheItems[i].IsUsed then
    begin
    	Result := i;	 // we found an open not used session
      exit;
    end;

  // if we reach this point there are no open and not used sessions
  Result := -1;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.NewSession
//
// Creates a new session. Tries to find a session in FChacheItems that
// is not open and reuses the entry. If none can be found a new entry
// in FCacheItems will be created.
// In case MaxSessions sessions already exists an exception is raised.
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
function TMGDBSessionCache.NewSession: Integer;
var
	i: Integer;
  DB: TDatabase;
  S: TSession;
begin
	// check if new sessions are allowed and if not raise an exception
  if not FAllowNewSessions then
  	raise ECannotCreateSession.Create('Not allowed to create new sessions');

	// check if we reached MaxSessions and in case raise an exception
  if FSessionsCount >= FMaxSessions then
  	raise Exception.Create('Maximum number of sessions reached');

	// create database component for new session
  DB := TDatabase.Create(Self);
  DB.AliasName := FDBAliasName;
  DB.ReadOnly := FDBReadOnly;
  DB.TransIsolation := FDBTransIsolation;
  DB.Params.Assign(FDBParams);
  DB.LoginPrompt := False;
  DB.KeepConnection := True;

	// iterate through cache searching for a not open session
	for i := 0 to Length(FCacheItems) - 1 do
  	if not FCacheItems[i].IsOpen then
    begin
    	// we found a not open entry and will reuse the cache entry
      // create new session and set properties
      S := Sessions.OpenSession(FSessionNamePrefix + IntToStr(i));
      S.PrivateDir := FSessionPrivateDir;
      S.NetFileDir := FSessionNetFileDir;

      DB.SessionName := S.SessionName;
      DB.DatabaseName := FDBNamePrefix + IntToStr(i);
    	FCacheItems[i].SetSession(DB, S);
			FSessionsCount := FSessionsCount + 1;
      DB.Open;
			Result := i;

      // call OnNewSession event
			EvOnNewSession;
      
    	exit;
    end;

  // if we reach this point there are no not open session so we'll have to
  // create a new one
  i := Length(FCacheItems);

  // create new session and set properties
  DB.DatabaseName := FDBNamePrefix + IntToStr(i);
  S := Sessions.OpenSession(FSessionNamePrefix + IntToStr(i));
  DB.SessionName := S.SessionName;
  S.PrivateDir := FSessionPrivateDir;
  S.NetFileDir := FSessionNetFileDir;
  SetLength(FCacheItems, i + 1);	// add space for one entry in cache
 	FCacheItems[i].SetSession(DB, S);
	FSessionsCount := FSessionsCount + 1;
  Result := i;

  // call OnNewSession event
	EvOnNewSession;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetOnCollectGarbage
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetOnCollectGarbage(const Value: TNotifyEvent);
begin
  FOnCollectGarbage := Value;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetOnNewSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetOnNewSession(const Value: TNotifyEvent);
begin
	FOnNewSession := Value;
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetOnRequestSession
//
// Author: Martin Geier
// Creation date: 2/6/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetOnRequestSession(const Value: TNotifyEvent);
begin
  FOnRequestSession := Value;
end;


//----------------------------------------------------------------------------------
// TMGDBSessionCache.SetOnReleaseSession
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.SetOnReleaseSession(const Value: TNotifyEvent);
begin
  FOnReleaseSession := Value;
end;

procedure Register;
begin
  RegisterComponents('MG', [TMGDBSessionCache]);
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.EvOnCollectGarbage
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.EvOnCollectGarbage;
begin
	if assigned(OnCollectGarbage) then
  	OnCollectGarbage(Self);
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.EvOnNewSession
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.EvOnNewSession;
begin
	if assigned(OnNewSession) then
  	OnNewSession(Self);
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.EvOnRequestSession
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.EvOnRequestSession;
begin
	if assigned(OnRequestSession) then
  	OnRequestSession(Self);
end;

//----------------------------------------------------------------------------------
// TMGDBSessionCache.EvOnReleaseSession
//
// Author: Martin Geier
// Creation date: 2/12/00
//----------------------------------------------------------------------------------
procedure TMGDBSessionCache.EvOnReleaseSession;
begin
	if assigned(OnReleaseSession) then
  	OnReleaseSession(Self);
end;

end.
